@@ -24,6 +24,7 @@
 #include "atk.h"
 #include "atkmarshal.h"
 #include "atk-enum-types.h"
+#include "atkintl.h"
 
 GPtrArray *extra_roles = NULL;
 
@@ -53,6 +54,7 @@ enum {
   PROPERTY_CHANGE,
   STATE_CHANGE,
   VISIBLE_DATA_CHANGED,
+  ACTIVE_DESCENDANT_CHANGED,
   
   LAST_SIGNAL
 };
@@ -183,6 +185,7 @@ atk_object_class_init (AtkObjectClass *klass)
   klass->focus_event = NULL;
   klass->property_change = NULL;
   klass->visible_data_changed = NULL;
+  klass->active_descendant_changed = NULL;
 
   g_object_class_install_property (gobject_class,
                                    PROP_NAME,
@@ -339,6 +342,15 @@ atk_object_class_init (AtkObjectClass *klass)
                   (GSignalAccumulator) NULL, NULL,
                   g_cclosure_marshal_VOID__VOID,
                   G_TYPE_NONE, 0);
+  atk_object_signals[ACTIVE_DESCENDANT_CHANGED] =
+    g_signal_new ("active_descendant_changed",
+		  G_TYPE_FROM_CLASS (klass),
+		  G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
+		  G_STRUCT_OFFSET (AtkObjectClass, active_descendant_changed),
+		  NULL, NULL,
+		  g_cclosure_marshal_VOID__POINTER,
+		  G_TYPE_NONE,
+		  1, G_TYPE_POINTER);
 }
 
 static void
@@ -1098,13 +1110,14 @@ atk_object_notify (GObject     *obj,
   g_signal_emit (obj, atk_object_signals[PROPERTY_CHANGE],
                  g_quark_from_string (pspec->name),
                  &values, NULL);
+  g_value_unset (&values.new_value);
 }
 
 /**
  * atk_role_get_name:
  * @role: The #AtkRole whose name is required
  *
- * Gets the description string describing the #Roleype @role.
+ * Gets the description string describing the #AtkRole @role.
  *
  * Returns: the string describing the AtkRole
  */
@@ -1140,6 +1153,261 @@ atk_role_get_name (AtkRole role)
   return name;
 }
 
+/**
+ * atk_role_get_localized_name:
+ * @role: The #AtkRole whose localized name is required
+ *
+ * Gets the localized description string describing the #AtkRole @role.
+ *
+ * Returns: the localized string describing the AtkRole
+ **/
+G_CONST_RETURN gchar*
+atk_role_get_localized_name (AtkRole role)
+{
+  G_CONST_RETURN gchar *name;
+  gboolean gettext_initialized = FALSE;
+
+#ifdef ENABLE_NLS
+  if (!gettext_initialized)
+    {
+      bindtextdomain (GETTEXT_PACKAGE, ATKLOCALEDIR);
+#ifdef HAVE_BIND_TEXTDOMAIN_CODESET
+      bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
+#endif
+    }
+#endif
+
+  switch (role)
+    {
+      case ATK_ROLE_INVALID:
+        name = _("invalid");
+        break;
+      case ATK_ROLE_ACCEL_LABEL:
+        name = _("accel-label");
+        break;
+      case ATK_ROLE_ALERT:
+        name = _("alert");
+        break;
+      case ATK_ROLE_ANIMATION:
+        name = _("animation");
+        break;
+      case ATK_ROLE_ARROW:
+        name = _("arrow");
+        break;
+      case ATK_ROLE_CALENDAR:
+        name = _("calendar");
+        break;
+      case ATK_ROLE_CANVAS:
+        name = _("canvas");
+        break;
+      case ATK_ROLE_CHECK_BOX:
+        name = _("check-box");
+        break;
+      case ATK_ROLE_CHECK_MENU_ITEM:
+        name = _("check-menu-item");
+        break;
+      case ATK_ROLE_COLOR_CHOOSER:
+        name = _("color-chooser");
+        break;
+      case ATK_ROLE_COLUMN_HEADER:
+        name = _("column-header");
+        break;
+      case ATK_ROLE_COMBO_BOX:
+        name = _("combo-box");
+        break;
+      case ATK_ROLE_DATE_EDITOR:
+        name = _("date-editor");
+        break;
+      case ATK_ROLE_DESKTOP_ICON:
+        name = _("desktop-icon");
+        break;
+      case ATK_ROLE_DESKTOP_FRAME:
+        name = _("desktop-frame");
+        break;
+      case ATK_ROLE_DIAL:
+        name = _("dial");
+        break;
+      case ATK_ROLE_DIALOG:
+        name = _("dialog");
+        break;
+      case ATK_ROLE_DIRECTORY_PANE:
+        name = _("directory-pane");
+        break;
+      case ATK_ROLE_DRAWING_AREA:
+        name = _("drawing-area");
+        break;
+      case ATK_ROLE_FILE_CHOOSER:
+        name = _("file-chooser");
+        break;
+      case ATK_ROLE_FILLER:
+        name = _("filler");
+        break;
+      case ATK_ROLE_FONT_CHOOSER:
+        name = _("font-chooser");
+        break;
+      case ATK_ROLE_FRAME:
+        name = _("frame");
+        break;
+      case ATK_ROLE_GLASS_PANE:
+        name = _("glass-pane");
+        break;
+      case ATK_ROLE_HTML_CONTAINER:
+        name = _("html-container");
+        break;
+      case ATK_ROLE_ICON:
+        name = _("icon");
+        break;
+      case ATK_ROLE_IMAGE:
+        name = _("image");
+        break;
+      case ATK_ROLE_INTERNAL_FRAME:
+        name = _("internal-frame");
+        break;
+      case ATK_ROLE_LABEL:
+        name = _("label");
+        break;
+      case ATK_ROLE_LAYERED_PANE:
+        name = _("layered-pane");
+        break;
+      case ATK_ROLE_LIST:
+        name = _("list");
+        break;
+      case ATK_ROLE_LIST_ITEM:
+        name = _("list-item");
+        break;
+      case ATK_ROLE_MENU:
+        name = _("menu");
+        break;
+      case ATK_ROLE_MENU_BAR:
+        name = _("menu-bar");
+        break;
+      case ATK_ROLE_MENU_ITEM:
+        name = _("menu-item");
+        break;
+      case ATK_ROLE_OPTION_PANE:
+        name = _("option-pane");
+        break;
+      case ATK_ROLE_PAGE_TAB:
+        name = _("page-tab");
+        break;
+      case ATK_ROLE_PAGE_TAB_LIST:
+        name = _("page-tab-list");
+        break;
+      case ATK_ROLE_PANEL:
+        name = _("panel");
+        break;
+      case ATK_ROLE_PASSWORD_TEXT:
+        name = _("password-text");
+        break;
+      case ATK_ROLE_POPUP_MENU:
+        name = _("popup-menu");
+        break;
+      case ATK_ROLE_PROGRESS_BAR:
+        name = _("progress-bar");
+        break;
+      case ATK_ROLE_PUSH_BUTTON:
+        name = _("push-button");
+        break;
+      case ATK_ROLE_RADIO_BUTTON:
+        name = _("radio-button");
+        break;
+      case ATK_ROLE_RADIO_MENU_ITEM:
+        name = _("radio-menu-item");
+        break;
+      case ATK_ROLE_ROOT_PANE:
+        name = _("root-pane");
+        break;
+      case ATK_ROLE_ROW_HEADER:
+        name = _("row-header");
+        break;
+      case ATK_ROLE_SCROLL_BAR:
+        name = _("scroll-bar");
+        break;
+      case ATK_ROLE_SCROLL_PANE:
+        name = _("scroll-pane");
+        break;
+      case ATK_ROLE_SEPARATOR:
+        name = _("separator");
+        break;
+      case ATK_ROLE_SLIDER:
+        name = _("slider");
+        break;
+      case ATK_ROLE_SPLIT_PANE:
+        name = _("split-pane");
+        break;
+      case ATK_ROLE_SPIN_BUTTON:
+        name = _("spin-button");
+        break;
+      case ATK_ROLE_STATUSBAR:
+        name = _("statusbar");
+        break;
+      case ATK_ROLE_TABLE:
+        name = _("table");
+        break;
+      case ATK_ROLE_TABLE_CELL:
+        name = _("table-cell");
+        break;
+      case ATK_ROLE_TABLE_COLUMN_HEADER:
+        name = _("table-column-header");
+        break;
+      case ATK_ROLE_TABLE_ROW_HEADER:
+        name = _("table-row-header");
+        break;
+      case ATK_ROLE_TEAR_OFF_MENU_ITEM:
+        name = _("tear-off-menu-item");
+        break;
+      case ATK_ROLE_TERMINAL:
+        name = _("terminal");
+        break;
+      case ATK_ROLE_TEXT:
+        name = _("text");
+        break;
+      case ATK_ROLE_TOGGLE_BUTTON:
+        name = _("toggle-button");
+        break;
+      case ATK_ROLE_TOOL_BAR:
+        name = _("tool-bar");
+        break;
+      case ATK_ROLE_TOOL_TIP:
+        name = _("tool-tip");
+        break;
+      case ATK_ROLE_TREE:
+        name = _("tree");
+        break;
+      case ATK_ROLE_TREE_TABLE:
+        name = _("tree-table");
+        break;
+      case ATK_ROLE_UNKNOWN:
+        name = _("unknown");
+        break;
+      case ATK_ROLE_VIEWPORT:
+        name = _("viewport");
+        break;
+      case ATK_ROLE_WINDOW:
+        name = _("window");
+        break;
+      case ATK_ROLE_HEADER:
+        name = _("header");
+        break;
+      case ATK_ROLE_FOOTER:
+        name = _("footer");
+        break;
+      case ATK_ROLE_PARAGRAPH:
+        name = _("paragraph");
+        break;
+      case ATK_ROLE_RULER:
+        name = _("ruler");
+        break;
+      case ATK_ROLE_APPLICATION:
+        name = _("application");
+        break;
+      default:
+        name = atk_role_get_name (role);
+        break;
+    }
+  return name;
+}
+
 /**
  * atk_role_for_name:
  * @name: a string which is the (non-localized) name of an ATK role.
@@ -1192,3 +1460,76 @@ atk_role_for_name (const gchar *name)
   
   return role;
 }
+
+/**
+ * atk_object_add_relationship:
+ * @object: The #AtkObject to which an AtkRelation is to be added. 
+ * @relationship: The #AtkRelationType of the relation
+ * @target: The #AtkObject which is to be the target of the relation.
+ *
+ * Adds a relationship of the specified type with the specified target.
+ *
+ * Returns TRUE if the relationship is added.
+ **/
+gboolean
+atk_object_add_relationship (AtkObject       *object,
+                             AtkRelationType relationship,
+                             AtkObject       *target)
+{
+  AtkObject *array[1];
+  AtkRelation *relation;
+
+  g_return_val_if_fail (ATK_IS_OBJECT (object), FALSE);
+  g_return_val_if_fail (ATK_IS_OBJECT (target), FALSE);
+
+  array[0] = target;
+  relation = atk_relation_new (array, 1, relationship);
+  atk_relation_set_add (object->relation_set, relation);
+  g_object_unref (relation);
+
+  return TRUE;
+}
+
+/**
+ * atk_object_remove_relationship:
+ * @object: The #AtkObject from which an AtkRelation is to be removed. 
+ * @relationship: The #AtkRelationType of the relation
+ * @target: The #AtkObject which is the target of the relation to be removed.
+ *
+ * Removes a relationship of the specified type with the specified target.
+ *
+ * Returns TRUE if the relationship is removed.
+ **/
+gboolean
+atk_object_remove_relationship (AtkObject       *object,
+                                AtkRelationType relationship,
+                                AtkObject       *target)
+{
+  gint n_relations, i;
+  gboolean ret = FALSE;
+  AtkRelation *relation;
+
+  g_return_val_if_fail (ATK_IS_OBJECT (object), FALSE);
+  g_return_val_if_fail (ATK_IS_OBJECT (target), FALSE);
+
+  n_relations = atk_relation_set_get_n_relations (object->relation_set);
+  for (i = 0; i < n_relations; i++)
+    {
+      relation = atk_relation_set_get_relation (object->relation_set, i);
+      if (atk_relation_get_relation_type (relation) == relationship)
+        {
+          GPtrArray *array;
+
+          array = atk_relation_get_target (relation);
+        
+          if (g_ptr_array_index (array, 0) == target)
+            {
+              atk_relation_set_remove (object->relation_set, relation); 
+              ret = TRUE;
+              break;
+            }
+        }
+    }
+
+  return ret;
+}
